#! /usr/bin/python3

import sys, os
# allow local modules like ressource_rc
sys.path.append(os.path.dirname(__file__))

from PyQt5.QtWidgets import QApplication, QMainWindow,QDialog,\
    QTableWidgetItem,QAbstractScrollArea,QPushButton, QWidget, \
    QFileDialog,QApplication
from PyQt5.QtCore import QTranslator, QLocale
from PyQt5.QtCore import QObject, pyqtSignal
from ui_clone_jbart import Ui_MainWindow
from ui_about import Ui_Dialog as Ui_AboutDialog
from dbus.mainloop.glib import DBusGMainLoop, threads_init
import gi
gi.require_version('UDisks', '2.0')
import usbDisk2
from monitor import Monitor
from subprocess import Popen, PIPE
from markdown import markdownFromFile
import io

class MyMain(QMainWindow):

    keysChanged  = pyqtSignal(str) # means that some USB stick has changed
    endmonitor   = pyqtSignal(Monitor)
    closemonitor = pyqtSignal(Monitor)
    
    def __init__(self, lesCles, owndisk, parent=None):
        """
        Main window for the application
        @param lesCles an instance of usbDisk2.Available
        @param owndisk an empty string or something like "sdc"
        @param parent parent window, defaults to None
        """
        QMainWindow.__init__(self, parent)
        self.lesCles=lesCles
        self.wd=os.path.abspath(os.path.dirname(__file__))
        self.monitors={}
        self.ui=Ui_MainWindow()
        self.ui.setupUi(self)
        self.initTable()
        self.initTab()
        self.initHelpTab()
        
        self.about=QDialog()
        self.about.ui=Ui_AboutDialog()
        self.about.ui.setupUi(self.about)
        self.initAbout()

        self.ui.action_About.triggered.connect(self.about.show)
        self.keysChanged.connect(self.update_keys)
        self.endmonitor.connect(self.monitor_finished)
        self.closemonitor.connect(self.monitor_closed)
        self.update_keys(owndisk) # check already connected USB sticks
        return

    def mdToEdit(self, path, edit):
        """
        Feeds a QTextEdit instance with a Markdown file
        @param path the path to the .md file, relative to this file
        @param edit a QTextEdit instance
        """
        path=os.path.join(self.wd, path)
        html=io.BytesIO()
        markdownFromFile(input=path, output=html)
        html.seek(0)
        edit.setHtml(html.read().decode("utf-8"))
        return        

    def initAbout(self):
        """
        Feeds the Dialog with a localized text
        """
        self.mdToEdit(self.tr("HelpAbout.md"), self.about.ui.textEdit)
        return
    
    def initHelpTab(self):
        """
        converts the Markdown help file and writes it into the help tab
        """
        self.mdToEdit(self.tr("HelpTab.md"), self.ui.textEdit)
        return
    
    def monitor_finished(self, monitor):
        """
        callback triggered when a monitor finished its job
        """
        t=self.ui.tableWidget
        t.setItem(monitor.rindex,4,
                  QTableWidgetItem(self.tr("Clone ready ... Tab #{0}").format(monitor.index)))
        t.resizeColumnsToContents()
        return
        
    def monitor_closed(self, monitor):
        """
        callback triggered when a monitor is closed
        """
        device = self.monitors.pop(monitor)
        w=self.ui.tabWidget.widget(monitor.index)
        self.ui.tabWidget.removeTab(monitor.index)
        w.close()
        t=self.ui.tableWidget
        t.setItem(monitor.rindex,4,
                  QTableWidgetItem(self.tr("Tab closed.")))
        t.resizeColumnsToContents()
        return

    def makeRow(self, rindex, disk, already):
        """
        Makes a row in the table view for a disk
        @param index of a row
        @param disk the disk instance
        @param already is True when the first colums is already correct,
        and the row already exists
        @param dic a dictionary disk => partitions
        @return a new value for the index, increased when a row is created
        """
        t=self.ui.tableWidget
        dic=self.lesCles.disksDict()
        nextIndex=rindex
        shortDisk=os.path.basename(disk)
        shortParts=sorted([os.path.basename(p) for p in dic[disk]])
        if not already:
            t.insertRow(rindex)
            t.setItem(rindex,0,QTableWidgetItem(shortDisk))
            nextIndex=rindex+1
            # create the button if it is not already there
            button=QPushButton(self.tr("Clone to {}").format(shortDisk))
            b_function=lambda:self.clonage("/dev/{}".format(shortDisk), button, rindex)
            button.clicked.connect(b_function)
            t.setCellWidget(rindex,3,button)
        t.setItem(rindex,1,QTableWidgetItem(", ".join(shortParts)))
        ## check whether there is still a Monitor with a process for this device
        devices=[os.path.basename(m.device) for  m in self.monitors]
        if shortDisk in devices:
            ## the disk is processed, do not enable the button
            button.setEnabled(False)
        return nextIndex

        

    def update_keys(self, owndisk):
        """
        callback function used when a USB stick is plugged in or off
        or when its partitins are modified.
        @param owndisk eventually points to the disk which was booted
        """
        # builds the list of USB sticks, except the stick enventually
        # used to boot the system
        disks=[d for d in  sorted(self.lesCles.disksDict())\
                   if not owndisk or owndisk not in d]
        t=self.ui.tableWidget
        rindex=0
        ## first, delete rows with disk references which are no longer valid
        for rindex in list(range(t.rowCount()))[::-1]:
            if t.itemAt(rindex,0).data(0) not in disks:
                t.removeRow(rindex)
        ## then, add new disks and update existing ones
        if disks:
            self.statusBar().showMessage(self.tr("Current USB disks: {}").format(", ".join([os.path.basename(d) for d in disks])))
        else:
            self.statusBar().showMessage(self.tr("No USB stick. Please plug a USB flash disk."))
        for disk in disks:
            shortDisk=os.path.basename(disk)
            while rindex < t.rowCount() and \
                  t.itemAt(rindex,0).data(0) < shortDisk:
                ## skip the lines before the place of shortDisk
                rindex+=1
            already = rindex < t.rowCount() and \
                t.itemAt(rindex,0).data(0) == shortDisk
            rindex=self.makeRow(rindex, disk, already)
        t.resizeColumnsToContents()
        return

    def initTable(self):
        """
        initialize the table of devices
        """
        t=self.ui.tableWidget
        t.setColumnCount(5)
        t.setHorizontalHeaderLabels([
            self.tr("Device"), self.tr("Partitions"),self.tr("Source"),
            self.tr("Actions"),self.tr("Status")
        ])
        t.setSizeAdjustPolicy(QAbstractScrollArea.AdjustToContents)
        return

    def initTab(self):
        """
        initialize the QTabWidget in the user interface
        removes the widget #1, which comes from Designer
        widget #0 is special: it contains some help
        """
        w=self.ui.tabWidget.widget(1)
        self.ui.tabWidget.removeTab(1)
        w.close()
        return

    def clonage(self,device, button, rindex):
        """
        Starts cloning
        @param device for example /dev/sdd
        @param button the button which was clicked
        @param rindex the row index of the button
        """
        # clonage can be triggered only once for a given device
        button.setEnabled(False)
        command="/usr/bin/python3"
        source=""
        if os.path.exists("/usr/lib/live/mount/persistence"):
            # allow only own-cloning when running from a live USB
            source="own"
        t=self.ui.tableWidget
        if source=="own":
            t.setItem(rindex,2,QTableWidgetItem(self.tr("Own running system")))
        else:
            # open a file dialog to get a source
            source, _ = QFileDialog.getOpenFileName(
                caption = self.tr("Choose an image to clone to {}").format(device),
                filter = self.tr("ISO Images (*.iso);;All files (*)")
            )
            if not source:
                return
            else:
                t.setItem(rindex,2,QTableWidgetItem(
                    "... "+os.path.dirname(source)[-10:]+"/"+os.path.basename(source)
                ))
                
        t.resizeColumnsToContents()
        args=[
            "{}/makeLiveStick.py".format(self.wd),
            "--no-persistence",
            "--source",
            source,
            device
        ]
        ### add a new tab and a monitor
        n=self.ui.tabWidget.count()
        # set a message in the table
        t=self.ui.tableWidget
        t.setItem(rindex,4,QTableWidgetItem(self.tr("Cloning ... Tab #{}").format(n)))
        t.resizeColumnsToContents()
        title=self.tr("Cloning to {}").format(device)
        newWidget=QWidget()
        self.ui.tabWidget.addTab(newWidget, title)
        m=Monitor(command, args, self, n, rindex, device)
        self.monitors[m]=device


def main():
    """
    The main call
    """
    app = QApplication(sys.argv)
    # i18n stuff
    locale = QLocale.system().name()
    translation="clone_jbart_{}.ts".format(locale)
    langPath=os.path.join(os.path.abspath(os.path.dirname(__file__)),"lang",translation)
    translator = QTranslator(app)
    translator.load(langPath)
    app.installTranslator(translator)
    #############
    lesCles=usbDisk2.Available()
    p=Popen("ls /usr/lib/live/mount/persistence", shell=True, stdout=PIPE, stderr=PIPE)
    out, err = p.communicate()
    shortDiskNames=out.decode("utf8").split()
    ownDisk=""
    if shortDiskNames:
        # if we have ['sda1', 'sda3'], this should give "sda"
        ownDisk=shortDiskNames[0][:-1]
    w = MyMain(lesCles, ownDisk)
    # addHook is not designed to work with a method inside an object
    # so, let us define the hook outside the main window, and relay
    # a signal to the main window
    # the slot keysChanged in MyMain accepts a string parameter
    # which is the current device if the system is booted from it
    def show_keys(man, obj):
        if lesCles.modified: # filter to detect only USB stick objects
            w.keysChanged.emit(ownDisk)
        lesCles.modified=False
        return
    lesCles.addHook('object-added',   show_keys)
    lesCles.addHook('object-removed', show_keys)
    
    w.show()
    
    sys.exit(app.exec_())

if __name__ == '__main__':
    main()
