import sys
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *
import datetime


class Monitor(QObject):
    def __init__(self, command, args, mainwindow, index, rindex, device):
        """
        the constructor
        @param command an executable
        @param args the command line args for the executable
        @param mainwindow a QWidget which should have signals endmonitor
        and closemonitor, and a QTabWidget in mainwindow.ui.tabWidget
        @param index the index of the wanted tab
        @param rindex the row index to access the table of devices
        @param device then name of the device targetted by the command
        """
        QObject.__init__(self)
        self.index=index
        self.rindex=rindex
        self.device=device
        self.w=mainwindow.ui.tabWidget.widget(index)
        mainwindow.ui.tabWidget.setCurrentIndex(index)
        self.mainwindow=mainwindow
        self.process = QProcess(self.w)
        self.process.readyReadStandardOutput.connect(self.readOutput)
        self.process.readyReadStandardError.connect(self.readErr)
        self.process.finished.connect(self.finish)
        self.stdoutTerminal = QPlainTextEdit(self.w)
        self.stderrTerminal = QPlainTextEdit(self.w)
        self.cursor=QTextCursor(self.stdoutTerminal.document());
        self.err_cursor=QTextCursor(self.stderrTerminal.document());
        self.layout = QVBoxLayout(self.w)
        self.layout.addWidget(QLabel(self.tr("stdout")))
        self.layout.addWidget(self.stdoutTerminal, stretch=4)
        self.layout.addWidget(QLabel(self.tr("stderr")))
        self.layout.addWidget(self.stderrTerminal, stretch=1)

        self.process.start(command, args)
        return

    def close(self):
        d=datetime.datetime.now()
        outfileName, ok = QInputDialog.getText(
            self.mainwindow,
            self.tr("Do you want to save the logs?"),
            self.tr("name of the file for the logs"),
            text="clone-log-{}.log".format(d.strftime("%Y-%m-%d--%H-%M"))
        )
        if ok:
            with open(outfileName, "w") as outfile:
                outfile.write("# STDOUT #\n")
                outfile.write(self.stdoutTerminal.toPlainText())
                outfile.write("\n# STDERR #\n")
                outfile.write(self.stderrTerminal.toPlainText())
        self.mainwindow.closemonitor.emit(self)
        return True
        
    def finish(self):
        button=QPushButton(self.tr("Cloning is over"))
        button.clicked.connect(self.close)
        self.layout.addWidget(button)
        self.mainwindow.endmonitor.emit(self)

    def read(self, source, cursor, edit):
        """
        reads a QByteArray from source and writes it into a QPlainTextEdit
        widget
        @param source: either an instance of QProcess.readAllStandardOutput
        or an instance of QProcess.readAllStandardError
        @param cursor a QTextCursor instance
        @param edit a QPlainTextEdit instance
        """
        text= bytes(source()).decode("utf-8")
        if text.startswith("\r"):
            ## carriage return
            cursor.select(cursor.LineUnderCursor)
            cursor.removeSelectedText()
            cursor.insertText(text[1:])
        else:
            cursor.insertText(text)
        edit.verticalScrollBar().setValue(edit.verticalScrollBar().maximum())
        return
    
    def readOutput(self):
        self.read(self.process.readAllStandardOutput,
                  self.cursor,
                  self.stdoutTerminal)
        return
    
    def readErr(self):
        self.read(self.process.readAllStandardError,
                  self.err_cursor,
                  self.stderrTerminal)
        return
    
if __name__ == "__main__":
    app = QApplication(sys.argv)
    main = Monitor('ls', ['-lR', '../..'])
    main.show()
    sys.exit(app.exec_())
